{
 "cells": [
  {
   "cell_type": "markdown",
   "id": "20b16754",
   "metadata": {},
   "source": [
    "# `elements.py`\n",
    "\n",
    "```{custom_download_link} elements.py\n",
    ":text: \".py\"\n",
    ":replace_default: \"False\"\n",
    "```\n",
    "\n",
    "```{custom_download_link} https://github.com/CIEM5000-2025/practice-assignments\n",
    ":text: \"All files practice assignments\"\n",
    ":replace_default: \"False\"\n",
    "```"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "2ffb5d3d",
   "metadata": {},
   "outputs": [],
   "source": [
    "import numpy as np\n",
    "import matplotlib.pyplot as plt"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "d60f78cd",
   "metadata": {},
   "outputs": [],
   "source": [
    "class Element:\n",
    "    \"\"\"\n",
    "    The Element class keeps track of each element in the model, including cross-section properties, \n",
    "    element orientation (for coordinate system transformations), and the nodes that make up each element. \n",
    "    With the help of the Node class, it also keeps track of which Degrees of Freedom (DOFs) belong to each element.\n",
    "\n",
    "    This class is responsible for providing the element stiffness matrix in the global coordinate system \n",
    "    (for subsequent assembly) and postprocessing element-level fields. \n",
    "\n",
    "    This class describes an element combining extension and Euler-Bernoulli bending. A similar (or inherited) \n",
    "    class could also be implemented for different element types (e.g., shear beam, Timoshenko beam, cable elements, etc). \n",
    "    For simplicity, it is assumed that elements are all arranged in a 2D plane.\n",
    "\n",
    "    Attributes:\n",
    "        node1 (Node): The first node of the element.\n",
    "        node2 (Node): The second node of the element.\n",
    "        EA (float): The axial stiffness of the element.\n",
    "        EI (float): The flexural stiffness of the element.\n",
    "\n",
    "    Methods:\n",
    "        clear(): Clears the counting of elements.\n",
    "        __init__(self, nodes): Initializes an Element object.\n",
    "        set_section(self, props): Sets the section properties of the element.\n",
    "        global_dofs(self): Returns the global degrees of freedom associated with the element.\n",
    "        stiffness(self): Calculate the stiffness matrix of the element.\n",
    "        add_distributed_load(self, q): Adds a distributed load to the element.\n",
    "        bending_moments(self, u_global, num_points=2): Calculate the bending moments along the element.\n",
    "        full_displacement(self, u_global, num_points=2): Calculates the displacement along the element.\n",
    "        plot_moment_diagram(self, u_elem, num_points=10, global_c=False, scale=1.0): Plots the bending moment diagram of the element.\n",
    "        plot_displaced(self, u_elem, num_points=10, global_c=False, scale=1.0): Plots the displaced element.\n",
    "        __str__(self): Returns a string representation of the Element object.\n",
    "    \"\"\"\n",
    "\n",
    "    ne = 0\n",
    "\n",
    "    def clear():\n",
    "        \"\"\"\n",
    "        Clears the counting of elements\n",
    "\n",
    "        This method resets the class-level counters for number of elements. \n",
    "        It should be used when you want to start a new problem from scratch.\n",
    "        \"\"\"\n",
    "        Element.ne = 0\n",
    "        \n",
    "    def __init__(self, node1, node2):\n",
    "        \"\"\"\n",
    "        Initializes an Element object.\n",
    "\n",
    "        Parameters:\n",
    "        - node1 (Node): The first node of the element.\n",
    "        - node2 (Node): The second node of the element.\n",
    "\n",
    "        Attributes:\n",
    "        - nodes (list): A list of Node objects representing the nodes of the element.\n",
    "        - L (float): Length of the element.\n",
    "        - cos (float): Cosine of the element's orientation angle.\n",
    "        - sin (float): Sine of the element's orientation angle.\n",
    "        - T (ndarray): Transformation matrix.\n",
    "        - Tt (ndarray): Transpose of the transformation matrix.\n",
    "\n",
    "        Returns:\n",
    "        None\n",
    "        \"\"\"\n",
    "        self.nodes = [node1, node2]\n",
    "\n",
    "        self.L = np.sqrt((self.nodes[1].x - self.nodes[0].x)**2.0 + (self.nodes[1].z - self.nodes[0].z)**2.0)\n",
    "\n",
    "        alpha = np.arctan2 #YOUR CODE HERE\n",
    "\n",
    "        T = np.zeros((6, 6))\n",
    "\n",
    "        T[0, 0] = T[1, 1] = T[3, 3] = T[4, 4] #YOUR CODE HERE\n",
    "        T[0, 1] = T[3, 4] #YOUR CODE HERE\n",
    "        T[1, 0] = T[4, 3] #YOUR CODE HERE\n",
    "        T[2, 2] = T[5, 5] #YOUR CODE HERE\n",
    "\n",
    "        self.T = T\n",
    "        self.Tt = np.transpose(T)\n",
    "\n",
    "        self.q = np.array([0,0])\n",
    "        self.local_element_load = np.array([0,0,0,0,0,0])\n",
    "        \n",
    "        Element.ne += 1\n",
    "\n",
    "    def set_section(self, props):\n",
    "        \"\"\"\n",
    "        Sets the section properties of the element.\n",
    "\n",
    "        Parameters:\n",
    "        - props (dict): A dictionary containing the section properties.\n",
    "                        The dictionary should have the following keys:\n",
    "                        - 'EA': The axial stiffness of the element.\n",
    "                        - 'EI': The flexural stiffness of the element.\n",
    "\n",
    "        Returns:\n",
    "        None\n",
    "        \"\"\"\n",
    "        if 'EA' in props:\n",
    "            self.EA = props['EA']\n",
    "        else:\n",
    "            self.EA = 1.e20\n",
    "        if 'EI' in props:\n",
    "            self.EI = props['EI']\n",
    "        else:\n",
    "            self.EI = 1.e20\n",
    "\n",
    "    def global_dofs(self):\n",
    "        \"\"\"\n",
    "        Returns the global degrees of freedom associated with the element.\n",
    "\n",
    "        Returns:\n",
    "            numpy.ndarray: Array containing the global degrees of freedom.\n",
    "        \"\"\"\n",
    "        return np.hstack((self.nodes[0].dofs, self.nodes[1].dofs))\n",
    "\n",
    "    def stiffness(self):\n",
    "        \"\"\"\n",
    "        Calculate the stiffness matrix of the element.\n",
    "\n",
    "        Returns:\n",
    "        np.ndarray: The stiffness matrix of the element.\n",
    "        \"\"\"\n",
    "        k = np.zeros((6, 6))\n",
    "\n",
    "        EA = self.EA\n",
    "        EI = self.EI\n",
    "        L = self.L\n",
    "\n",
    "        #YOUR CODE HERE\n",
    "\n",
    "        return np.matmul(np.matmul(self.Tt, k), self.T)\n",
    "\n",
    "    def add_distributed_load(self, q):\n",
    "        \"\"\"\n",
    "        Adds a distributed load to the element.\n",
    "\n",
    "        Parameters:\n",
    "            q (list): List of distributed load in local x and z direction.\n",
    "\n",
    "        Returns:\n",
    "            None\n",
    "        \"\"\"\n",
    "\n",
    "        l = self.L\n",
    "        self.q = np.array(q)\n",
    "\n",
    "        self.local_element_load # =[YOUR CODE HERE, , , , , ]\n",
    "\n",
    "        global_element_load #YOUR CODE HERE\n",
    "\n",
    "        self.nodes[0].add_load #YOUR CODE HERE\n",
    "        self.nodes[1].add_load #YOUR CODE HERE\n",
    "\n",
    "    def bending_moments(self, u_global, num_points=2):\n",
    "        \"\"\"\n",
    "        Calculate the bending moments along the element.\n",
    "\n",
    "        Parameters:\n",
    "        - u_global (numpy.ndarray): Global displacement vector.\n",
    "        - num_points (int): Number of points to evaluate the bending moments. Default is 2.\n",
    "\n",
    "        Returns:\n",
    "        - M (numpy.ndarray): Array of bending moments at the specified points.\n",
    "        \"\"\"\n",
    "\n",
    "        l = self.L\n",
    "        q = self.q[1]\n",
    "        EI = self.EI\n",
    "\n",
    "        local_x = np.linspace(0.0, l, num_points)\n",
    "\n",
    "        local_disp #YOUR CODE HERE\n",
    "\n",
    "        M #YOUR CODE HERE\n",
    "        \n",
    "        return M\n",
    "    \n",
    "    def full_displacement (self, u_global, num_points=2):\n",
    "        \"\"\"\n",
    "        Calculates the displacement along the element.\n",
    "\n",
    "        Args:\n",
    "            u_global (numpy.ndarray): Global displacement vector of the element.\n",
    "            num_points (int, optional): Number of points to calculate the bending moments. Default is 2.\n",
    "\n",
    "        Returns:\n",
    "            numpy.ndarray: Array of displacement along the element.\n",
    "        \"\"\"\n",
    "        #YOUR CODE HERE\n",
    "\n",
    "        u #YOUR CODE HERE\n",
    "        w #YOUR CODE HERE\n",
    "\n",
    "        return u, w\n",
    "    \n",
    "    def plot_moment_diagram (self, u_elem, num_points=10, global_c=False, scale=1.0):\n",
    "        \"\"\"\n",
    "        Plots the bending moment diagram of the element.\n",
    "\n",
    "        Args:\n",
    "            u_global (numpy.ndarray): Global displacement vector of the element.\n",
    "            num_points (int, optional): Number of points to calculate the bending moments. Default is 2.\n",
    "            global_c (bool, optional): If True, plots the bending moment diagram in the global coordinate system. Default is False (plots in local coordinate system).\n",
    "            scale (float, optional): Scale factor for the bending moment diagram. Default is 1.0.\n",
    "\n",
    "        Returns:\n",
    "            None\n",
    "        \"\"\"\n",
    "        import matplotlib.pyplot as plt\n",
    "\n",
    "        x = np.linspace ( 0.0, self.L, num_points )\n",
    "        M = self.bending_moments ( u_elem, num_points )\n",
    "        xM_local = np.vstack((np.hstack([0,x,x[-1]]),np.hstack([0,M,0])*scale))\n",
    "        if global_c:\n",
    "            xM_global = np.matmul(self.Tt[0:2,:2],xM_local)\n",
    "            xz_start_node = np.vstack((np.ones(num_points+2)*self.nodes[0].x, np.ones(num_points+2)*self.nodes[0].z))\n",
    "            xz_Mlijn = xM_global + xz_start_node\n",
    "            p = plt.plot(xz_Mlijn[0,:],xz_Mlijn[1,:])\n",
    "            X0= self.nodes[0].x\n",
    "            Z0= self.nodes[0].z\n",
    "            X1= self.nodes[1].x\n",
    "            Z1= self.nodes[1].z\n",
    "            plt.plot((X0, X1), (Z0, Z1), color=p[0].get_color())\n",
    "            plt.axis('off')\n",
    "            plt.axis('equal')\n",
    "        else:\n",
    "            p = plt.plot(xM_local[0,:],xM_local[1,:])\n",
    "            plt.xlabel ( \"x\" )\n",
    "            plt.ylabel ( \"M\" )\n",
    "        if not plt.gca().yaxis_inverted():\n",
    "            plt.gca().invert_yaxis()\n",
    "        plt.gcf().patch.set_alpha(0.0)\n",
    "        plt.gca().patch.set_alpha(0.0)\n",
    "        plt.gca().patch.set_alpha(0.0)\n",
    "        plt.title('Moment line')\n",
    "\n",
    "    def plot_displaced(self, u_elem, num_points=10, global_c=False, scale=1.0):\n",
    "        \"\"\"\n",
    "        Plots the displacd element.\n",
    "\n",
    "        Args:\n",
    "            u_global (numpy.ndarray): Global displacement vector of the element.\n",
    "            num_points (int, optional): Number of points to calculate the bending moments. Default is 2.\n",
    "            global_c (bool, optional): If True, plots the displacement diagram in the global coordinate system. Default is False (plots in local coordinate system).\n",
    "            scale (float, optional): Scale factor for the displacement diagram. Default is 1.0.\n",
    "\n",
    "        Returns:\n",
    "            None\n",
    "        \"\"\"\n",
    "\n",
    "        x = np.linspace ( 0.0, self.L, num_points )\n",
    "        u, w = self.full_displacement ( u_elem, num_points )\n",
    "        uw_local = np.vstack((x+u*scale,w*scale))\n",
    "        if global_c:\n",
    "            uw_global = np.matmul(self.Tt[:2,:2],uw_local)\n",
    "            xz_start_node = np.vstack((np.ones(num_points)*self.nodes[0].x, np.ones(num_points)*self.nodes[0].z))\n",
    "            uw = uw_global + xz_start_node\n",
    "            p =  plt.plot(uw[0,:],uw[1,:])\n",
    "            X0= self.nodes[0].x\n",
    "            Z0= self.nodes[0].z\n",
    "            X1= self.nodes[1].x\n",
    "            Z1= self.nodes[1].z\n",
    "            plt.plot((X0, X1), (Z0, Z1), color=p[0].get_color(),alpha=0.3)\n",
    "            plt.axis('off')\n",
    "            plt.axis('equal')\n",
    "        else:\n",
    "            p = plt.plot(uw_local[0,:],uw_local[1,:])\n",
    "            plt.plot((0, self.L), (0, 0), color=p[0].get_color(),alpha=0.3)\n",
    "        if not plt.gca().yaxis_inverted():\n",
    "            plt.gca().invert_yaxis()\n",
    "        plt.gcf().patch.set_alpha(0.0)\n",
    "        plt.gca().patch.set_alpha(0.0)\n",
    "        plt.gca().patch.set_alpha(0.0)\n",
    "        plt.title('Displaced structure')\n",
    "\n",
    "    def plot_numbered_structure(self,beam_number):\n",
    "        \"\"\"\n",
    "        Plots the nodes and elements of the structure with their node and element numbers.\n",
    "\n",
    "        Returns:\n",
    "            None\n",
    "        \"\"\"\n",
    "\n",
    "        X0= self.nodes[0].x\n",
    "        Z0= self.nodes[0].z\n",
    "        X1= self.nodes[1].x\n",
    "        Z1= self.nodes[1].z\n",
    "        node_num = []\n",
    "        node_num.append(self.nodes[0].dofs[0] // 3)\n",
    "        node_num.append(self.nodes[1].dofs[0] // 3)\n",
    "        plt.plot((X0, X1), (Z0, Z1), color='black',alpha=0.3)\n",
    "        for i, node in enumerate(self.nodes):\n",
    "            plt.text(node.x, node.z, f'[{node.dofs[0] // 3}]', fontsize=12, ha='center', va='center')\n",
    "        plt.text((X0+X1)/2, (Z0+Z1)/2, f'({beam_number})', fontsize=12, ha='center', va='center')\n",
    "        if not plt.gca().yaxis_inverted():\n",
    "            plt.gca().invert_yaxis()\n",
    "        plt.axis('off')\n",
    "        plt.axis('equal')\n",
    "        plt.gcf().patch.set_alpha(0.0)\n",
    "        plt.gca().patch.set_alpha(0.0)\n",
    "        plt.gca().patch.set_alpha(0.0)\n",
    "\n",
    "\n",
    "    def __str__(self):\n",
    "        \"\"\"\n",
    "        Returns a string representation of the Element object.\n",
    "        \n",
    "        The string includes the values of the node1, node2 attributes.\n",
    "        \"\"\"\n",
    "        return f\"Element connecting:\\nnode #1:\\n {self.nodes[0]}\\nwith node #2:\\n {self.nodes[1]}\""
   ]
  }
 ],
 "metadata": {
  "jupytext": {
   "text_representation": {
    "extension": ".md",
    "format_name": "myst",
    "format_version": 0.13,
    "jupytext_version": "1.16.2"
   }
  },
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "source_map": [
   12,
   26,
   31
  ]
 },
 "nbformat": 4,
 "nbformat_minor": 5
}